""" This enables easy access to indian states and union territories """
# @coding: utf-8
# @author: Rishabh Batra
# @email: rishabhbatra1002@gmail.com

# python imports
import csv
from typing import List, AnyStr

# module imports
from . import utils

STATES = []
UNION_TERRITORIES = []
STATES_AND_TERRITORIES = []

_lookup_cache = {}


class State:
    """ defines a state"""
    def __init__(self, abbr: str, name: str, capital: str, population: int, area: int, lang: str):
        self.name = name
        self.abbr = abbr
        self.capital = capital
        self.population = self.stats_validation(population)
        self.area = self.stats_validation(area)
        self.density = self.calc_population_density(
            self.population,
            self.area
        )
        self.language = lang

    def __repr__(self):
        return "<State: {}>".format(self.name)

    def __str__(self):
        return self.name

    @staticmethod
    def stats_validation(stats: (str, int, None)=None) -> (int, None):
        """Validating Statistics entered and converting them to string"""

        if not isinstance(stats, (str, int, type(None))):
            raise TypeError(
                "Stats should either be a string of integers, integer or None. Provided type {}"
                .format(type(stats)))

        stats_is_not_none = stats is not None
        if stats_is_not_none:
            stats = int(stats)
        else:
            stats = None
        return stats

    @staticmethod
    def calc_population_density(population: int, area: int) -> (int, None):
        """ Calculating the Population Density given the Population and Area """
        if population is not None and area is not None:
            return population // area


#  the class functions of the package ends here #
def load_states():
    """
    Load state data from a csv file in the package

    Creates list of states.
    Also adds state abbrevation attribute access to the package.
    # in.states.AR
    """

    with open(utils.FILE_NAME['states']) as statesfile:
        states = csv.reader(statesfile, delimiter=',')
        for row in states:
            state = State(abbr=row[0],
                          name=row[1],
                          capital=row[2],
                          population=row[3],
                          area=row[4],
                          lang=row[5]
                          )

            # creates list of states
            STATES.append(state)

            # creates a master list of all states and territories
            STATES_AND_TERRITORIES.append(state)

            # provide package-level abbrevation
            globals()[state.abbr] = state


def load_ut():
    """
    Load union territories from a csv file provided with the package.

    Creates a list of Union Teriitories.
    Also add abbrevation attribute access to the package
    # in.states.PY

    """
    with open(utils.FILE_NAME['ut']) as utsfile:
        uts = csv.reader(utsfile, delimiter=',')
        for row in uts:
            ut = State(abbr=row[0],
                       name=row[1],
                       capital=row[2],
                       population=row[3],
                       area=row[4],
                       lang=row[5]
                       )

            # creates a list of uts
            UNION_TERRITORIES.append(ut)

            # appends to the master list of all states and territories
            STATES_AND_TERRITORIES.append(ut)

            # provides package-level abbrevation
            globals()[ut.abbr] = ut


def lookup(val: str, field: str=None, use_cache: bool=True) -> str:
    """ This provides functionality to find state based on the
    lookup value provided.

    :param str val: if val has 2 alphas it will look for abbrevation
                    anything else will try to match the state names

    :param str field: can take values, 'None', 'abbr', 'name' to bypass fuzzy matching.

    :param bool use_cache: Caching of the Loaded States
    This method caches non-None results, but cache can be bypassed with the
    `use_cache=False` argument.
    TODO: add another attribute to this for lookup of metaphones
        """
    # jellyfish for fuzzy matching
    # capitalizing the value to standardize formatting

    if field is None:
        if utils.ABBR_RE.match(val):
            val = val.upper()
            field = 'abbr'
        else:
            val = val.capitalize()
            field = 'name'
    elif field == 'name':
        val = val.capitalize()

    # see if result is in cache
    cache_key = "%s:%s" % (field, val)
    if use_cache and cache_key in _lookup_cache:
        return _lookup_cache[cache_key]

    for state in STATES_AND_TERRITORIES:
        if val == getattr(state, field):
            _lookup_cache[cache_key] = state
            return state

    raise ValueError('{}. matching state not found. Please try entering something else.'.format(val))


def get_states(entity: str = 'all') -> List[AnyStr]:
    """ This returns a list of strings of entities sorted in ascending order if entity param is not passed
    :param entity: enum with allowed values ('state', 'ut', 'all')
           default: 'all'
    Example:
    >>> from india import states
    >>> print(states.get_states())
    """
    allowed_entities = ('state', 'ut', 'all')
    # validation
    if entity not in allowed_entities:
        raise ValueError(f'Wrong entity type passed: {entity}. Allowed values {allowed_entities}')

    state_names = []

    # appending state names
    if entity in ('all', 'state'):
        for state in STATES:
            state_names.append(state.name)

    # appending ut names
    if entity in ('all', 'ut'):
        for ut in UNION_TERRITORIES:
            state_names.append(ut.name)

    return sorted(state_names)


# initialising states and territories
load_states()
load_ut()



""" This enables easy access to indian districts in the respective states """
# @coding: utf-8
# @author: Rishabh Batra
# @email: rishabhbatra1002@gmail.com
# Get weather for every city using weather API

# python imports
import csv
from typing import List, AnyStr

# module imports
from . import config, states, utils

CITIES = []

_lookup_cache = {}


class City:
    """
    Defines a city
    """
    def __init__(self,
                 abbr: str,
                 name: str,
                 state: str,
                 population: int=None,
                 area: int=None,
                 url: str=None):
        self.name = name
        self.state = states.lookup(state)
        self.abbr = abbr
        self.state_abbr = '%s_%s' % (self.state.abbr, abbr)
        self.population = states.State.stats_validation(population)
        self.area = states.State.stats_validation(area)
        self.density = states.State.calc_population_density(
            self.population, self.area
        )
        self.url = url

    def __repr__(self):
        return "<City: %s>" % self.name

    def __str__(self):
        return self.name

    def is_capital(self):
        """
        :return: boolean whether the city is capital or not
        """
        return bool(self.name == self.state.capital)

    @staticmethod
    def weather():
        """
        :return:
        """
        config.Config().get('Machine')


def load_city():
    """
    Loads districs of india data from csv file.

    Also adds city level abbreviation to the package
    # in.cities.DL_CD
    :return:
    """
    with open(utils.FILE_NAME['cities'], encoding="utf-8") as citiesfile:
        cities = csv.reader(citiesfile, delimiter=',')
        next(cities)
        for row in cities:
            # reassigning empty strings of population area or url
            if len(row[4]) == 0:
                row[4] = None
            if len(row[5]) == 0:
                row[5] = None
            if len(row[7]) == 0:
                row[7] = None

            city = City(state=row[0],
                        abbr=row[1],
                        name=row[2],
                        population=row[4],
                        area=row[5],
                        url=row[7]
                        )
            # list of cities
            CITIES.append(city)

            globals()[city.state_abbr] = city

    return 0


def lookup(val: str, field: str=None, use_cache: bool=True):
    """ This provides functionality to find city based on the
       lookup value provided.

       :param str val: if val has 2 alphas it will look for abbrevation
                       anything else will try to match the state names

       :param str field: can take values, 'None', 'abbr', 'name' to bypass fuzzy matching.

       :param bool use_cache:
       This method caches non-None results, but cache can be bypassed with the
       `use_cache=False` argument.

       TODO: Add fuzzy matching
    """
    if field is None:
        if utils.ABBR_RE.match(val):
            val = val.upper()
            field = 'abbr'
        else:
            val = val.capitalize()
            field = 'name'

    # check if result in cahe
    cache_key = "%s:%s" %(field, val)
    if use_cache and cache_key in _lookup_cache:
        return _lookup_cache[cache_key]

    for city in CITIES:
        if val == getattr(city, field):
            _lookup_cache[cache_key] = city
            return city


def get_cities() -> List[AnyStr]:
    """ This returns a list of strings of cities sorted in ascending order
    Example:
    >>> from india import cities
    >>> print(cities.get_cities())
    """

    city_names = []
    for city in CITIES:
        city_names.append(city.name)
    return sorted(city_names)


# initialising cities
load_city()



""" Config parsing to create a singleton """
# @coding: utf-8
# @author: Rishabh Batra
# @email: rishabhbatra1002@gmail.com

# external imports
from configparser import ConfigParser


class Config:
    """
    Handles the singleton use
    of the application configuration
    """
    class __ActualConfig:
        def __init__(self, path):
            self.config = ConfigParser()
            with open(path) as infl:
                self.config.read_file(infl)
            self.validate()

        def validate(self):
            """ Validating the Config file to check for required attributes """
            # Validating whether Required Configurations are available
            if not self.config.has_option('APIXU_API', 'KEY'):
                raise ImportWarning('''Authorisation Key not found in config file \n
                                     Not able to fetch the Weather data. \
                                     Please add KEY to the config file.''')
            if not self.config.has_option('APIXU_API', 'BASE_URL'):
                raise ImportWarning('''Base Url not found for Support API \n
                                       Not able to fetch the Weather data. \
                                       Please add KEY to the config file.''')

    config_instance = None

    def __init__(self, path=None):
        if not Config.config_instance and path:
            Config.config_instance = Config.__ActualConfig(path)

    def __getattr__(self, name):
        """Proxy to the singleton configuration object
        """
        return getattr(self.config_instance.config, name)

    def get_items(self, section):
        """
        Get all the items found in the section
        :param section:
        :return:
        """
        return dict(self.config_instance.config.items(section))



"""Initializing module India """
# @coding: utf-8
# @author: Rishabh Batra
# @email: rishabhbatra1002@gmail.com

# Module imports
from . import states, cities
from .states import STATES, UNION_TERRITORIES, STATES_AND_TERRITORIES
from .cities import CITIES

__appname__ = 'pyindia'
__version__ = '1.1.1'



""" Utils File supporting the module"""
# @coding: utf-8
# @author: Rishabh Batra
# @email: rishabhbatra1002@gmail.com
# python imports
import os
import re

WORKING_DIRECTORY = os.path.dirname(os.path.realpath(__file__))

FILE_NAME = {
    'states': WORKING_DIRECTORY + '/data/states.csv',
    'ut': WORKING_DIRECTORY + '/data/ut.csv',
    'cities': WORKING_DIRECTORY + '/data/cities.csv',
    'towns': WORKING_DIRECTORY + '/data/towns.csv',
    'config': WORKING_DIRECTORY + '/data/in.conf'
}

ABBR_RE = re.compile(r'^[a-zA-Z]{2,3}$')



""" This enables access to indian towns in respective cities """
# @coding: utf-8
# @author: Rishabh Batra
# @email: rishabhbatra1002@gmail.com
# Write class Town and creating a Town object for every town in database

# python imports

# module imports
from . import cities

TOWNS = []


class Town:
    """
    Defines a Town
    """
    def __init__(self,
                 pin_code: int,
                 name: str,
                 city: str,
                 latitude: float,
                 longitude: float,
                 accuracy: float):

        self.pin_code = pin_code
        self.name = name
        self.city = cities.lookup(city)
        self.state = self.city.state
        self.latitude = latitude
        self.longitude = longitude
        self.accuracy = accuracy

    def __repr__(self):
        return '<Town: %s>' % self.name

    def __str__(self):
        return self.name
